/*
 * Copyright (c) 2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include "functional.cuh"
#include <raft/linalg/unary_op.cuh>

namespace raft {
namespace linalg {
namespace detail {

template <typename math_t, typename IdxType = int>
void divideScalar(math_t* out, const math_t* in, math_t scalar, IdxType len, cudaStream_t stream)
{
  raft::linalg::unaryOp(out, in, len, divides_scalar<math_t>(scalar), stream);
}

};  // end namespace detail
};  // end namespace linalg
};  // end namespace raft
