
/*
 * Copyright (c) 2020-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <iostream>
#include <rmm/device_uvector.hpp>

namespace raft {
namespace mst {
namespace detail {

template <typename idx_t>
__device__ idx_t get_1D_idx()
{
  return blockIdx.x * blockDim.x + threadIdx.x;
}

}  // namespace detail
}  // namespace mst
}  // namespace raft
