/*
 * Copyright (c) 2021-2022, NVIDIA CORPORATION.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <raft/distance/specializations/detail/canberra.cuh>
#include <raft/distance/specializations/detail/chebyshev.cuh>
#include <raft/distance/specializations/detail/correlation.cuh>
#include <raft/distance/specializations/detail/cosine.cuh>
#include <raft/distance/specializations/detail/hamming_unexpanded.cuh>
#include <raft/distance/specializations/detail/hellinger_expanded.cuh>
#include <raft/distance/specializations/detail/jensen_shannon.cuh>
#include <raft/distance/specializations/detail/kl_divergence.cuh>
#include <raft/distance/specializations/detail/l1.cuh>
#include <raft/distance/specializations/detail/l2_expanded.cuh>
#include <raft/distance/specializations/detail/l2_sqrt_expanded.cuh>
#include <raft/distance/specializations/detail/l2_sqrt_unexpanded.cuh>
#include <raft/distance/specializations/detail/l2_unexpanded.cuh>
#include <raft/distance/specializations/detail/lp_unexpanded.cuh>
